==> wwwoffle-1.2c.tgz: wwwoffle-1.2c/wwwoffled.c <==
/***************************************
  $Header: /home/amb/wwwoffle/RCS/wwwoffled.c 1.18 1997/07/06 09:18:57 amb Exp $

  WWWOFFLE - World Wide Web Offline Explorer - Version 1.2c.
  A demon program to maintain the database and spawn the servers.
  ******************/ /******************
  Written by Andrew M. Bishop

  This file Copyright 1996,97 Andrew M. Bishop
  It may be distributed under the GNU Public License, version 2, or
  any higher version.  See section COPYING of the GNU Public license
  for conditions under which this file may be redistributed.
  ***************************************/


#include <stdio.h>
#include <stdlib.h>

#include <sys/wait.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>

#include "wwwoffle.h"
#include "config.h"
#include "sockets.h"
#include "errors.h"


static void usage(int verbose);
static void demoninit(void);
static void sigchild(int signum);
static void sigexit(int signum);
static void sighup(int signum);


/*+ The server sockets that we listen on +*/
int http_fd=-1,                 /*+ for the HTTP connections. +*/
    wwwoffle_fd=-1;             /*+ for the WWWOFFLE connections. +*/

/*+ The online / offline status. +*/
int online=0;

/*+ The current number active servers +*/
int n_servers=0,                /*+ in total. +*/
    n_fetch_servers=0;          /*+ fetching a page. +*/

/*+ The wwwoffle client file descriptor when fetching. +*/
int fetch_fd=-1;

/*+ The pids of the servers that are fetching. +*/
int fetch_pids[MAX_FETCH_SERVERS];

/*+ The current status, fetching or not. +*/
int fetching=0;


/*+ Set to 1 when the demon is to shutdown. +*/
static int closedown=0;

/*+ Set to 1 when the demon is to re-read config file due to a sighup. +*/
static int readconfig=0;

/*+ The pid of a child that has exited. +*/
static int exit_pid[MAX_SERVERS];

/*+ The status of the child that has exited. +*/
static int exit_status[MAX_SERVERS];

/*+ True if run as a demon +*/
static int detached=1;


/*++++++++++++++++++++++++++++++++++++++
  The main program.
  ++++++++++++++++++++++++++++++++++++++*/

int main(int argc, char** argv)
{
 int i;
 int err;
 struct stat buf;

 /* Parse the command line options */

 for(i=1;i<argc;i++)
   {
    if(!strcmp(argv[i],"-h"))
       usage(1);

    if(!strcmp(argv[i],"-d"))
      {
       detached=0;
       continue;
      }
   
    if(!strcmp(argv[i],"-c"))
      {
       if(++i>=argc)
         {fprintf(stderr,"wwwoffled: The '-c' option requires a filename.\n"); exit(1);}

       ConfigFile=argv[i];
       continue;
      }

    fprintf(stderr,"wwwoffled: Unknown option '%s'.\n",argv[i]); exit(1);
   }

 /* Initialise things. */

 for(i=0;i<MAX_FETCH_SERVERS;i++)
    fetch_pids[i]=0;

 for(i=0;i<MAX_SERVERS;i++)
    exit_pid[i]=exit_status[i]=0;

 InitErrorHandler("wwwoffled",1,!detached);

 if(ReadConfigFile(2))
    PrintMessage(Fatal,"Error in configuration file '%s'.",ConfigFile);

 PrintMessage(Important,"WWWOFFLE Demon Version 1.2c started");
 PrintMessage(Inform,"WWWOFFLE Read Configuration File.");

 if(WWWOFFLE_Gid != -1)
   {
    if(setgid(WWWOFFLE_Gid)<0)
       PrintMessage(Fatal,"Cannot set group id to %d [%!s].",WWWOFFLE_Gid);
    if(setegid(WWWOFFLE_Gid)<0)
       PrintMessage(Fatal,"Cannot set effective group id to %d [%!s].",WWWOFFLE_Gid);
   }

 if(WWWOFFLE_Uid != -1)
   {
    if(setuid(WWWOFFLE_Uid)<0)
       PrintMessage(Fatal,"Cannot set user id to %d [%!s].",WWWOFFLE_Uid);
    if(seteuid(WWWOFFLE_Uid)<0)
       PrintMessage(Fatal,"Cannot set effective user id to %d [%!s].",WWWOFFLE_Uid);
   }

 if(WWWOFFLE_Gid != -1 || WWWOFFLE_Uid != -1)
    PrintMessage(Inform,"Running with uid=%d, gid=%d.",getuid(),getgid());
    
 http_fd=OpenServerSocket(HTTP_Port);
 if(http_fd==-1)
    PrintMessage(Fatal,"Cannot create HTTP server socket.");

 wwwoffle_fd=OpenServerSocket(WWWOFFLE_Port);
 if(wwwoffle_fd==-1)
    PrintMessage(Fatal,"Cannot create WWWOFFLE server socket.");

 if(stat(SpoolDir,&buf))
   {
    err=mkdir(SpoolDir,0755);
    if(err==-1 && errno!=EEXIST)
       PrintMessage(Fatal,"Cannot create spool directory %s [%!s].",SpoolDir);
    stat(SpoolDir,&buf);
   }

 if(!S_ISDIR(buf.st_mode))
    PrintMessage(Fatal,"The spool directory %s is not a directory.",SpoolDir);

 err=chdir(SpoolDir);
 if(err==-1)
    PrintMessage(Fatal,"Cannot change to spool directory %s [%!s].",SpoolDir);

 if(detached)
   {
    demoninit();
    PrintMessage(Important,"Detached from terminal and changed pid to %d.",getpid());
    InitErrorHandler("wwwoffled",1,!detached); /* pid changes after detaching. */
   }

 sigchild(0);
 sigexit(0);
 sighup(0);

 /* Loop around waiting for connections. */

 do
   {
    struct timeval tv;
    fd_set readfd;
    int nfds;
    int retval;

    if(http_fd>wwwoffle_fd)
       nfds=http_fd+1;
    else
       nfds=wwwoffle_fd+1;

    FD_ZERO(&readfd);

    FD_SET(wwwoffle_fd,&readfd);

    if(n_servers<MAX_SERVERS)
       FD_SET(http_fd,&readfd);

    tv.tv_sec=10;
    tv.tv_usec=0;

    retval=select(nfds,&readfd,NULL,NULL,&tv);

    if(retval!=-1)
      {
       if(FD_ISSET(wwwoffle_fd,&readfd))
         {
          char *host;
          int port,client;

          client=AcceptConnect(wwwoffle_fd,&host,&port);
          if(client>=0)
            {
             if(IsAllowedConnect(host))
               {
                FILE *file=fdopen(client,"r");

                PrintMessage(Important,"WWWOFFLE connection from %s.",host);

                if(!file)
                  {
                   PrintMessage(Warning,"Unable to fdopen the wwwoffle control socket [%!s].");
                   CloseSocket(client);
                  }
                else
                  {
                   CommandConnect(file);

                   if(fetch_fd!=client)
                      fclose(file);
                  }
               }
             else
               {
                PrintMessage(Warning,"WWWOFFLE connection from %s rejected.",host);
                CloseSocket(client);
               }
            }
         }

       if(FD_ISSET(http_fd,&readfd))
         {
          char *host;
          int port,client;

          client=AcceptConnect(http_fd,&host,&port);
          if(client>=0)
            {
             if(IsAllowedConnect(host))
               {
                PrintMessage(Inform,"HTTP Proxy connection from %s.",host);
                ForkServer(client,1);
               }
             else
                PrintMessage(Warning,"HTTP Proxy connection from %s rejected.",host);

             CloseSocket(client);
            }
         }
      }

    if(readconfig)
      {
       readconfig=0;

       PrintMessage(Important,"SIGHUP signalled.");

       PrintMessage(Important,"WWWOFFLE Re-reading Configuration File.");

       if(ReadConfigFile(-1))
          PrintMessage(Warning,"Error in configuration file; keeping old values.");

       PrintMessage(Important,"WWWOFFLE Finished Re-reading Configuration File.");
      }

    for(i=0;i<MAX_SERVERS;i++)
       if(exit_pid[i]!=0)
         {
          if(exit_status[i]>=0)
             PrintMessage(Inform,"Child wwwoffles exited with status %d (pid=%d).",exit_status[i],exit_pid[i]);
          else
             PrintMessage(Important,"Child wwwoffles terminated by signal %d (pid=%d).",-exit_status[i],exit_pid[i]);
          exit_pid[i]=exit_status[i]=0;
         }

    if(fetch_fd!=-1)
      {
       if(!online)
          fetching=0;

       while(fetching && n_fetch_servers<MAX_FETCH_SERVERS && n_servers<MAX_SERVERS)
          ForkServer(fetch_fd,0);

       if(!fetching && n_fetch_servers==0)
         {
          char *msg="WWWOFFLE No more to fetch.\n";
          write(fetch_fd,msg,strlen(msg));
          close(fetch_fd);
          fetch_fd=-1;

          PrintMessage(Important,"WWWOFFLE Fetch finished.");
         }
      }
   }
 while(!closedown);

 PrintMessage(Important,"Exit signalled.");

 /* Close down and exit. */

 CloseSocket(http_fd);
 CloseSocket(wwwoffle_fd);

 return(0);
}


/*++++++++++++++++++++++++++++++++++++++
  Print the program usage in long or short format.

  int verbose True for long format.
  ++++++++++++++++++++++++++++++++++++++*/

static void usage(int verbose)
{
 fputs("\n"
       "WWWOFFLED - World Wide Web Offline Explorer (Demon) - Version 1.2c\n"
       "\n",stderr);

 if(verbose)
    fputs("(c) Andrew M. Bishop 1996,97 [       amb@gedanken.demon.co.uk ]\n"
          "                             [http://www.gedanken.demon.co.uk/]\n"
          "\n",stderr);

 fputs("Usage: wwwoffled [-h] [-c <config-file>] [-d]\n"
       "\n",stderr);

 if(verbose)
    fprintf(stderr,
            "   -h                  : Display this help.\n"
            "\n"
            "   -c <config-file>    : The name of the configuration file to use.\n"
            "                         (Default %s).\n"
            "\n"
            "   -d                  : Do not detach from the terminal and use stderr.\n"
            "\n",ConfigFile);

 exit(0);
}


/*++++++++++++++++++++++++++++++++++++++
  Detach ourself from the controlling terminal.
  ++++++++++++++++++++++++++++++++++++++*/

static void demoninit(void)
{
 pid_t pid=0;

 pid=fork();
 if(pid==-1)
    PrintMessage(Fatal,"Cannot fork() to detach(1) [%!s].");
 else if(pid)
    exit(0);

 setsid();

 pid=fork();
 if(pid==-1)
    PrintMessage(Fatal,"Cannot fork() to detach(2) [%!s].");
 else if(pid)
    exit(0);

 /*Close fds*/
 close(STDIN_FILENO);
 close(STDOUT_FILENO);
 close(STDERR_FILENO);
}


/*++++++++++++++++++++++++++++++++++++++
  The signal handler for the child exiting.

  int signum The signal number.
  ++++++++++++++++++++++++++++++++++++++*/

static void sigchild(int signum)
{
 void (*sig)();
 int pid;
 int status;

 if(signum)
    while((pid=waitpid(-1,&status,WNOHANG))>0)
      {
       int i;
       int exitval=0;

       if(WIFEXITED(status))
          exitval=WEXITSTATUS(status);
       else if(WIFSIGNALED(status))
          exitval=-WTERMSIG(status);

       for(i=0;i<MAX_SERVERS;i++)
          if(exit_pid[i]==0)
            {
             exit_pid[i]=pid;
             exit_status[i]=exitval;
             break;
            }

       n_servers--;

       if(fetch_fd!=-1)
         {
          for(i=0;i<MAX_FETCH_SERVERS;i++)
             if(fetch_pids[i]==pid)
                break;

          if(i!=MAX_FETCH_SERVERS)
            {
             n_fetch_servers--;
             fetch_pids[i]=0;

             if(exitval==3)
                fetching=0;

             if(exitval==4 && online)
                fetching=1;
            }
         }
      }

#if defined(SIGCHLD)
 sig=signal(SIGCHLD,sigchild);
 if(sig==SIG_ERR)
    PrintMessage(Fatal,"Cannot add SIGGHLD handler.");
#else
 sig=signal(SIGCLD,sigchild);
 if(sig==SIG_ERR)
    PrintMessage(Fatal,"Cannot add SIGGLD handler.");
#endif
}


/*++++++++++++++++++++++++++++++++++++++
  The signal handler for the signals to tell us to exit.

  int signum The signal number.
  ++++++++++++++++++++++++++++++++++++++*/

static void sigexit(int signum)
{
 void (*sig)();

 if(signum)
    closedown=1;

 sig=signal(SIGINT,sigexit);
 if(sig==SIG_ERR)
    PrintMessage(Warning,"Cannot add SIGINT handler.\n");

 sig=signal(SIGQUIT,sigexit);
 if(sig==SIG_ERR)
    PrintMessage(Warning,"Cannot add SIGQUIT handler.\n");

 sig=signal(SIGTERM,sigexit);
 if(sig==SIG_ERR)
    PrintMessage(Warning,"Cannot add SIGTERM handler.\n");
}


/*++++++++++++++++++++++++++++++++++++++
  The signal handler for the signals to tell us to re-read the config file.

  int signum The signal number.
  ++++++++++++++++++++++++++++++++++++++*/

static void sighup(int signum)
{
 void (*sig)();

 if(signum)
    readconfig=1;

 sig=signal(SIGHUP,sighup);
 if(sig==SIG_ERR)
    PrintMessage(Warning,"Cannot add SIGHUP handler.\n");
}

==> wwwoffle-2.0c.tgz: wwwoffle-2.0c/wwwoffled.c <==
/***************************************
  $Header: /home/amb/wwwoffle/RCS/wwwoffled.c 2.6 1997/12/27 12:07:52 amb Exp $

  WWWOFFLE - World Wide Web Offline Explorer - Version 2.0a.
  A demon program to maintain the database and spawn the servers.
  ******************/ /******************
  Written by Andrew M. Bishop

  This file Copyright 1996,97 Andrew M. Bishop
  It may be distributed under the GNU Public License, version 2, or
  any higher version.  See section COPYING of the GNU Public license
  for conditions under which this file may be redistributed.
  ***************************************/


#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include <sys/wait.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>

#include "wwwoffle.h"
#include "misc.h"
#include "config.h"
#include "sockets.h"
#include "errors.h"


static void usage(int verbose);
static void demoninit(void);
static void install_sighandlers(void);
static void sigchild(int signum);
static void sigexit(int signum);
static void sighup(int signum);


/*+ The server sockets that we listen on +*/
int http_fd=-1,                 /*+ for the HTTP connections. +*/
    wwwoffle_fd=-1;             /*+ for the WWWOFFLE connections. +*/

/*+ The online / offline status. +*/
int online=0;

/*+ The current number active servers +*/
int n_servers=0,                /*+ in total. +*/
    n_fetch_servers=0;          /*+ fetching a page. +*/

/*+ The wwwoffle client file descriptor when fetching. +*/
int fetch_fd=-1;

/*+ The pids of the servers that are fetching. +*/
int fetch_pids[MAX_FETCH_SERVERS];

/*+ The current status, fetching or not. +*/
int fetching=0;

/*+ Set to 1 when the demon is to shutdown. +*/
static int closedown=0;

/*+ Set to 1 when the demon is to re-read config file due to a sighup. +*/
static int readconfig=0;

/*+ Remember that we got a SIGCHLD +*/
static sig_atomic_t got_sigchld=0;

/*+ True if run as a demon +*/
static int detached=1;


/*++++++++++++++++++++++++++++++++++++++
  The main program.
  ++++++++++++++++++++++++++++++++++++++*/

int main(int argc, char** argv)
{
 int i;
 int err;
 struct stat buf;

 /* Parse the command line options */

 for(i=1;i<argc;i++)
   {
    if(!strcmp(argv[i],"-h"))
       usage(1);

    if(!strcmp(argv[i],"-d"))
      {
       detached=0;
       if(i<(argc-1) && isdigit(argv[i+1][0]) && !argv[i+1][1])
         {
          DebugLevel=Fatal+1-atoi(argv[++i]);
          if(DebugLevel<0)
             DebugLevel=0;
         }
       continue;
      }

    if(!strcmp(argv[i],"-c"))
      {
       if(++i>=argc)
         {fprintf(stderr,"wwwoffled: The '-c' option requires a filename.\n"); exit(1);}

       ConfigFile=argv[i];
       continue;
      }

    fprintf(stderr,"wwwoffled: Unknown option '%s'.\n",argv[i]); exit(1);
   }

 /* Initialise things. */

 for(i=0;i<MAX_FETCH_SERVERS;i++)
    fetch_pids[i]=0;

 InitErrorHandler("wwwoffled",1,!detached);

 if(ReadConfigFile(2))
    PrintMessage(Fatal,"Error in configuration file '%s'.",ConfigFile);

 PrintMessage(Important,"WWWOFFLE Demon Version 2.0 started");
 PrintMessage(Inform,"WWWOFFLE Read Configuration File.");

 if(WWWOFFLE_Gid != -1)
   {
    if(setgid(WWWOFFLE_Gid)<0)
       PrintMessage(Fatal,"Cannot set group id to %d [%!s].",WWWOFFLE_Gid);
    if(setegid(WWWOFFLE_Gid)<0)
       PrintMessage(Fatal,"Cannot set effective group id to %d [%!s].",WWWOFFLE_Gid);
   }

 if(WWWOFFLE_Uid != -1)
   {
    if(setuid(WWWOFFLE_Uid)<0)
       PrintMessage(Fatal,"Cannot set user id to %d [%!s].",WWWOFFLE_Uid);
    if(seteuid(WWWOFFLE_Uid)<0)
       PrintMessage(Fatal,"Cannot set effective user id to %d [%!s].",WWWOFFLE_Uid);
   }

 if(WWWOFFLE_Gid != -1 || WWWOFFLE_Uid != -1)
    PrintMessage(Inform,"Running with uid=%d, gid=%d.",getuid(),getgid());

 http_fd=OpenServerSocket(HTTP_Port);
 if(http_fd==-1)
    PrintMessage(Fatal,"Cannot create HTTP server socket.");

 wwwoffle_fd=OpenServerSocket(WWWOFFLE_Port);
 if(wwwoffle_fd==-1)
    PrintMessage(Fatal,"Cannot create WWWOFFLE server socket.");

 if(stat(SpoolDir,&buf))
   {
    err=mkdir(SpoolDir,0755);
    if(err==-1 && errno!=EEXIST)
       PrintMessage(Fatal,"Cannot create spool directory %s [%!s].",SpoolDir);
    stat(SpoolDir,&buf);
   }

 if(!S_ISDIR(buf.st_mode))
    PrintMessage(Fatal,"The spool directory %s is not a directory.",SpoolDir);

 err=chdir(SpoolDir);
 if(err==-1)
    PrintMessage(Fatal,"Cannot change to spool directory %s [%!s].",SpoolDir);

 if(detached)
   {
    demoninit();
    PrintMessage(Important,"Detached from terminal and changed pid to %d.",getpid());
    InitErrorHandler("wwwoffled",1,!detached); /* pid changes after detaching. */
   }

 install_sighandlers();

 /* Loop around waiting for connections. */

 do
   {
    struct timeval tv;
    fd_set readfd;
    int nfds;
    int retval;

    if(http_fd>wwwoffle_fd)
       nfds=http_fd+1;
    else
       nfds=wwwoffle_fd+1;

    FD_ZERO(&readfd);

    FD_SET(wwwoffle_fd,&readfd);

    if(n_servers<MaxServers)
       FD_SET(http_fd,&readfd);

    tv.tv_sec=10;
    tv.tv_usec=0;

    retval=select(nfds,&readfd,NULL,NULL,&tv);

    if(retval!=-1)
      {
       if(FD_ISSET(wwwoffle_fd,&readfd))
         {
          char *host;
          int port,client;

          client=AcceptConnect(wwwoffle_fd);
          init_buffer(client);

          if(client>=0 && !SocketRemoteName(client,&host,NULL,&port))
            {
             if(IsAllowedConnect(host))
               {
                PrintMessage(Important,"WWWOFFLE Connection from %s.",host);

                CommandConnect(client);

                if(fetch_fd!=client)
                   CloseSocket(client);
               }
             else
               {
                PrintMessage(Warning,"WWWOFFLE connection from %s rejected.",host);
                CloseSocket(client);
               }
            }
         }

       if(FD_ISSET(http_fd,&readfd))
         {
          char *host;
          int port,client;

          client=AcceptConnect(http_fd);
          init_buffer(client);

          if(client>=0 && !SocketRemoteName(client,&host,NULL,&port))
            {
             if(IsAllowedConnect(host))
               {
                PrintMessage(Inform,"HTTP Proxy connection from %s.",host);
                ForkServer(client,1);
               }
             else
                PrintMessage(Warning,"HTTP Proxy connection from %s rejected.",host);

             CloseSocket(client);
            }
         }
      }

    if(readconfig)
      {
       readconfig=0;

       PrintMessage(Important,"SIGHUP signalled.");

       PrintMessage(Important,"WWWOFFLE Re-reading Configuration File.");

       if(ReadConfigFile(-1))
          PrintMessage(Warning,"Error in configuration file; keeping old values.");

       PrintMessage(Important,"WWWOFFLE Finished Re-reading Configuration File.");
      }

    if(got_sigchld)
      {
       int pid, status;

       /* To avoid race conditions, reset the flag before fetching the status */
       got_sigchld=0;

       while((pid=waitpid(-1,&status,WNOHANG))>0)
        {
         int i;
         int exitval=0;

         if(WIFEXITED(status))
            exitval=WEXITSTATUS(status);
         else if(WIFSIGNALED(status))
            exitval=-WTERMSIG(status);

         if(exitval>=0)
            PrintMessage(Inform,"Child wwwoffles exited with status %d (pid=%d).",exitval,pid);
         else
            PrintMessage(Important,"Child wwwoffles terminated by signal %d (pid=%d).",-exitval,pid);

         n_servers--;

         /* Check if the child that terminated is one of the fetching wwwoffles */
         for(i=0;i<MaxFetchServers;i++)
            if(fetch_pids[i]==pid)
              {
               n_fetch_servers--;
               fetch_pids[i]=0;
               break;
              }

         if(exitval==3)
            fetching=0;

         if(exitval==4 && online)
            fetching=1;

         if(!online)
            fetching=0;
        }

        PrintMessage(Debug,"Currently running: %d servers total, %d fetchers.",n_servers,n_fetch_servers);
      }

    /* The select timed out or we got a signal. If we are currently fetching,
       start fetch servers to look for jobs in the spool directory. */

    while(fetching && n_fetch_servers<MaxFetchServers && n_servers<MaxServers)
       ForkServer(fetch_fd,0);

    if(fetch_fd!=-1 && !fetching && n_fetch_servers==0)
      {
       write_string(fetch_fd,"WWWOFFLE No more to fetch.\n");
       CloseSocket(fetch_fd);
       fetch_fd=-1;

       PrintMessage(Important,"WWWOFFLE Fetch finished.");
      }
   }
 while(!closedown);

 /* Close down and exit. */

 CloseSocket(http_fd);
 CloseSocket(wwwoffle_fd);

 if(n_servers)
    PrintMessage(Important,"Exit signalled - waiting for children.");
 else
    PrintMessage(Important,"Exit signalled.");

 while(n_servers)
   {
    int pid,status,exitval=0;

    while((pid=waitpid(-1,&status,0))>0)
      {
       if(WIFEXITED(status))
          exitval=WEXITSTATUS(status);
       else if(WIFSIGNALED(status))
          exitval=-WTERMSIG(status);

       if(exitval>=0)
          PrintMessage(Inform,"Child wwwoffles exited with status %d (pid=%d).",exitval,pid);
       else
          PrintMessage(Important,"Child wwwoffles terminated by signal %d (pid=%d).",-exitval,pid);

       n_servers--;
      }
   }

 PrintMessage(Important,"Exiting.");

 return(0);
}


/*++++++++++++++++++++++++++++++++++++++
  Print the program usage in long or short format.

  int verbose True for long format.
  ++++++++++++++++++++++++++++++++++++++*/

static void usage(int verbose)
{
 fputs("\n"
       "WWWOFFLED - World Wide Web Offline Explorer (Demon) - Version 2.0a\n"
       "\n",stderr);

 if(verbose)
    fputs("(c) Andrew M. Bishop 1996,97 [       amb@gedanken.demon.co.uk ]\n"
          "                             [http://www.gedanken.demon.co.uk/]\n"
          "\n",stderr);

 fputs("Usage: wwwoffled [-h] [-c <config-file>] [-d [<log-level>]]\n"
       "\n",stderr);

 if(verbose)
    fprintf(stderr,
            "   -h                  : Display this help.\n"
            "\n"
            "   -c <config-file>    : The name of the configuration file to use.\n"
            "                         (Default %s).\n"
            "\n"
            "   -d [<log-level>]    : Do not detach from the terminal and use stderr.\n"
            "                         The log-level is a number from 0 to 5.\n"
            "\n",ConfigFile);

 exit(0);
}


/*++++++++++++++++++++++++++++++++++++++
  Detach ourself from the controlling terminal.
  ++++++++++++++++++++++++++++++++++++++*/

static void demoninit(void)
{
 pid_t pid=0;

 pid=fork();
 if(pid==-1)
    PrintMessage(Fatal,"Cannot fork() to detach(1) [%!s].");
 else if(pid)
    exit(0);

 setsid();

 pid=fork();
 if(pid==-1)
    PrintMessage(Fatal,"Cannot fork() to detach(2) [%!s].");
 else if(pid)
    exit(0);

 /* Close fds */
 close(STDIN_FILENO);
 close(STDOUT_FILENO);
 close(STDERR_FILENO);
}


/*++++++++++++++++++++++++++++++++++++++
  Install the signal handlers.
  ++++++++++++++++++++++++++++++++++++++*/

static void install_sighandlers(void)
{
 struct sigaction action;

 /* SIGCHLD */
 action.sa_handler = sigchild;
 sigemptyset (&action.sa_mask);
 action.sa_flags = 0;
 if(sigaction(SIGCHLD, &action, NULL) != 0)
    PrintMessage(Warning, "Cannot install SIGCHLD handler.");

 /* SIGINT, SIGQUIT, SIGTERM */
 action.sa_handler = sigexit;
 sigemptyset(&action.sa_mask);
 sigaddset(&action.sa_mask, SIGINT);           /* Block all of them */
 sigaddset(&action.sa_mask, SIGQUIT);
 sigaddset(&action.sa_mask, SIGTERM);
 action.sa_flags = 0;
 if(sigaction(SIGINT, &action, NULL) != 0)
    PrintMessage(Warning, "Cannot install SIGINT handler.");
 if(sigaction(SIGQUIT, &action, NULL) != 0)
    PrintMessage(Warning, "Cannot install SIGQUIT handler.");
 if(sigaction(SIGTERM, &action, NULL) != 0)
    PrintMessage(Warning, "Cannot install SIGTERM handler.");

 /* SIGHUP */
 action.sa_handler = sighup;
 sigemptyset(&action.sa_mask);
 action.sa_flags = 0;
 if(sigaction(SIGHUP, &action, NULL) != 0)
    PrintMessage(Warning, "Cannot install SIGHUP handler.");
}


/*++++++++++++++++++++++++++++++++++++++
  The signal handler for the child exiting.

  int signum The signal number.
  ++++++++++++++++++++++++++++++++++++++*/

static void sigchild(int signum)
{
 got_sigchld=1;
}


/*++++++++++++++++++++++++++++++++++++++
  The signal handler for the signals to tell us to exit.

  int signum The signal number.
  ++++++++++++++++++++++++++++++++++++++*/

static void sigexit(int signum)
{
 closedown=1;
}


/*++++++++++++++++++++++++++++++++++++++
  The signal handler for the signals to tell us to re-read the config file.

  int signum The signal number.
  ++++++++++++++++++++++++++++++++++++++*/

static void sighup(int signum)
{
 readconfig=1;
}
